'''
Entry point of the library for website backends and CLI applications.
'''
from .main import Main


def main():
    ticker = input('Enter the ticker: ')
    main_obj = Main(ticker)


if __name__ == '__main__':
    main()



'''
This module helps in plotting the various details related to the model and the predictions, 
This module requires the processed data from the pre-processing module, training data from the train module,
Market information from the market module, and the calculations from the calculations module. 
'''
import os  
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  
import matplotlib.pyplot as plt
from . import train, preprocessing, market, calculations


class Plots:
    def __init__(self, ticker):
        self.preprocessing = preprocessing.data_preprocessing(ticker)
        self.train = train.Training(ticker)
        self.market = market.Market(ticker)
        self.calculations = calculations.Calculations()
    '''
    This function will plot the training data of the model and will use loss as the metric to depict the quality of the model. 
    This can be analsysed by TensorBoard as well, as tensorflow is a dependency of the stockDL library, 
    TensorBoard provides comparison of traning sessions by comparing the traget metrics after changing the values of parameters and hyperparameters.  
    '''
    def plot_training_data(self, model_selected, metric='loss', val_metric='val_loss'):
        plt.plot(model_selected.history[metric])
        plt.plot(model_selected.history[val_metric])
        plt.title('model loss')
        plt.ylabel('loss')
        plt.xlabel('epoch')
        plt.legend(['train', 'test'], loc='upper right')
        plt.show()
    
    '''
    This function plots the actual financial data of a stock and the predictions by the two deep learning strategies used in the library.
    This is a sanity check for the model and will give an idea to the user about the accuracy of the models. 
    '''
    def plot_predictions(self):
        plt.figure(figsize=(30, 10))
        plt.plot(self.preprocessing.y_train, label="Actual")
        plt.plot(self.train.y_pred_train_lstm, label="Prediction by LSTM Model")
        plt.plot(self.train.y_pred_train_mix, label="Prediction by Mix-LSTM Model")
        plt.legend(fontsize=20)
        plt.grid(axis="both")
        plt.title("Actual Open Price and Pedicted Ones on train set", fontsize=20)
        plt.show()

    '''
    This function plots the months we would stay in the market for trading as 1's and the months we would stay out of the market as 0's.
    The plot would overlap the predictions plot thus giving the user a visualisation if it was a good choice to stay out of the market for the particular month.

    Plotting the In-Out Months: 
        In: 1 we stay in the market as the price is greater than the actual price of the stock. 
        Out: 0 we stay out of the market as the price of the stock is less than the actual price of the stock. 
    '''

    def in_out(self):
        plt.figure(figsize=(30, 10))
        plt.plot(self.preprocessing.y_test, label="Actual")
        plt.plot(self.train.y_pred_lstm, label="LSTM Predictions")
        plt.plot(self.market.v_lstm, label="In and out LSTM")
        plt.plot(self.train.y_pred_mix, label="Mix LSTM Predictions")
        plt.plot(self.market.v_mix, label="In and out Mix LSTM")
        plt.legend(fontsize=20)
        plt.grid(axis="both")
        plt.title(
            "Actual Open Price, Predicted Ones and Vectors on In and Out Moments", fontsize=20)
        plt.show()

    '''
    This plotting function plots a comparison plot between the traditional financial strategies and the deep learning strategies,
    thus giving the user freedom to understand and decide what strategy to go with. 
    '''
    def comparison_plots(self):
        plt.figure(figsize=(30, 10))
        plt.plot(self.calculations.gross_portfolio(self.preprocessing.test, self.preprocessing.v_bh), label="Portfolio Buy and Hold Strategy")
        plt.plot(self.calculations.gross_portfolio(self.preprocessing.test, self.preprocessing.v_ma),
                label="Portfolio Moving Average Strategy")
        plt.plot(self.calculations.gross_portfolio(self.preprocessing.test, self.market.v_lstm), label="Portfolio LSTM Model")
        plt.plot(self.calculations.gross_portfolio(self.preprocessing.test, self.market.v_mix), label="Portfolio Mix LSTM Model")
        plt.legend(fontsize=20)
        plt.grid(axis="both")
        plt.title("Gross Portfolios of three methods", fontsize=20)
        plt.show()



'''
This module handles the training of the two-deep learning strategies used in this library.
It requires preprocessing and models modules and their dependencies. 
'''
import os  
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  
from keras.callbacks import ReduceLROnPlateau
from . import models
from . import preprocessing


class Training:
    def __init__(self, ticker):
        self.preprocessing = preprocessing.data_preprocessing(ticker)
        self.models = models.Models(ticker)
        
        '''Prevents false minima by reducing the learning rates on plateaus. '''

        self.learning_rate_reduction = ReduceLROnPlateau(monitor='val_loss',
                                                         patience=25,
                                                         verbose=1,
                                                         factor=0.25,
                                                         min_lr=0.00001)
        '''Uncomment and add tensorboard to callbacks to run TensorBoard for visualisation
        self.tensorboard = keras.callbacks.TensorBoard(
                                                        log_dir='TensorBoard_Logs',
                                                        histogram_freq=1,
                                                        embeddings_freq=1
                                                    )  
        '''
        # self.lstm_history, self.mix_history = self.train_model()
        # self.models.lstm_model.save_weights(lstm_weights.h5")
        # self.models.mix_lstm_model.save_weights("mix_lstm_weights.h5")
        self.y_pred_train_lstm = self.models.lstm_model.predict(self.preprocessing.X_train)
        self.y_pred_train_mix = self.models.mix_lstm_model.predict(self.preprocessing.X_train)
        self.y_pred_lstm = self.models.lstm_model.predict(self.preprocessing.X_test)
        self.y_pred_mix = self.models.mix_lstm_model.predict(self.preprocessing.X_test)

    '''Currently both models are trained together by this function. '''
    def train_model(self):
        '''Stores the history of the LSTM model. '''
        lstm_history = self.models.lstm_model.fit(self.preprocessing.X_train, self.preprocessing.y_train, epochs=400,
                                                  batch_size=48, validation_data=(self.preprocessing.X_test,
                                                                                  self.preprocessing.y_test),
                                                  verbose=1, callbacks=[self.learning_rate_reduction], shuffle=False)
        '''Stores the history of the Conv1D + LSTM Model or the Mix Model. '''
        mix_history = self.models.mix_lstm_model.fit(self.preprocessing.X_train, self.preprocessing.y_train, epochs=400,
                                                     batch_size=48, validation_data=(self.preprocessing.X_test,
                                                                                     self.preprocessing.y_test),
                                                     verbose=1, callbacks=[self.learning_rate_reduction], shuffle=False)
        return lstm_history, mix_history



'''
This is the final module responsible for the result calculation and parsing the dataframe to JSON.
This module requires data from preprocessing, calculations, market and all their dependencies.
'''
import os  
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  
import pandas as pd
from . import calculations, preprocessing, market


class Results:
    def __init__(self, ticker):
        self.preprocessing = preprocessing.data_preprocessing(ticker)
        self.market = market.Market(ticker)
        self.calculations = calculations.Calculations()
        self.result = self.result_calculations()

    '''
    This function handles the result calculation based on the formula devised in the calculation module,
    It generates the total yield in a 6-year investment and annual yields as well.
    '''
    def result_calculations(self):
        print("Test period of {:.2f} years, from {} to {} \n".format(len(self.preprocessing.v_bh) / 12,
                                                                     str(self.preprocessing.test.loc[
                                                                             self.preprocessing.test.index[0],
                                                                             "First Day Current Month"])[:10],
                                                                     str(self.preprocessing.test.loc[
                                                                             self.preprocessing.test.index[-1],
                                                                             "First Day Next Month"])[:10]))

        results = pd.DataFrame({})
        results["Method"] = ["Buy and hold", "Moving average", "LSTM", "Mix"]
        ''' This list stores the market data predictions obtained from the market module. '''
        vs = [self.preprocessing.v_bh, self.preprocessing.v_ma, self.market.v_lstm, self.market.v_mix]
        results["Total Gross Yield"] = [
            str(round(self.calculations.gross_yield(self.preprocessing.test, vi)[0], 2)) + " %" for vi in vs]
        results["Annual Gross Yield"] = [
            str(round(self.calculations.gross_yield(self.preprocessing.test, vi)[1], 2)) + " %" for vi in vs]
        results["Total Net Yield"] = [str(round(self.calculations.net_yield(self.preprocessing.test, vi)[0], 2)) + " %"
                                      for vi in vs]
        results["Annual Net Yield"] = [str(round(self.calculations.net_yield(self.preprocessing.test, vi)[1], 2)) + " %"
                                       for vi in vs]

        return results



'''
This module makes the data ready for predictions and calculations,
This module requires the data module to get the data from the Yahoo finance API.
This module has 3 main components:
1. Convert the daily stock data to monthly data for analysis based on the opening prices for the month. 
2. Creating an analysis window for predictions. [6 years is the window size in stockDL]
3. Making data RNN ready.
'''
import os  
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  
import numpy as np
from sklearn.preprocessing import MinMaxScaler
import pandas as pd
from . import data


class data_preprocessing:
    def __init__(self, ticker):
        self.data_reader = data.Data_Loader(ticker)
        self.df_monthly = self.monthly_df(self.data_reader.df)
        self.window = 5
        self.X, self.y = self.data_scaling(self.df_monthly)
        self.split = 72
        self.X_train = self.X[:-self.split-1, :, :]
        self.X_test = self.X[-self.split-1:, :, :]
        self.y_train = self.y[:-self.split-1]
        self.y_test = self.y[-self.split-1:]
        self.test = self.df_monthly.iloc[-self.split:, :]
        self.v_bh = np.ones(self.test.shape[0])
        self.v_ma = self.test["First Day Current Month Opening"] > self.test["mv_avg_12"]
    '''
    This function converts the daily stock data to montly by considering the first day of the month as the day to decide
    if we should stay in the market or out. 
    It also creates moving average of the data for 12 and 24 months to device more insights for traditional investment stratergy.
    '''
    def monthly_df(self, df):

        dfm = df.resample("M").mean()
        dfm = dfm[:-1]

        dfm["First Day Current Month"] = self.data_reader.first_days[:-1]
        dfm["First Day Next Month"] = self.data_reader.first_days[1:]
        dfm["First Day Current Month Opening"] = np.array(
            df.loc[self.data_reader.first_days[:-1], "Open"])
        dfm["First Day Next Month Opening"] = np.array(
            df.loc[self.data_reader.first_days[1:], "Open"])
        dfm["Quotient"] = dfm["First Day Next Month Opening"].divide(
            dfm["First Day Current Month Opening"])

        dfm["mv_avg_12"] = dfm["Open"].rolling(window=12).mean().shift(1)
        dfm["mv_avg_24"] = dfm["Open"].rolling(window=24).mean().shift(1)

        dfm = dfm.iloc[24:, :]
        return dfm

    '''
    This function creates a sliding window for feeding the financial data to the DL model for training and analysis,
    For stockDL we use a window of 6 past years from the current date/ year. 
    '''
    def create_window(self, data, window_size):
        data_s = data.copy()
        for i in range(window_size):
            data = pd.concat([data, data_s.shift(-(i + 1))], axis=1)

        data.dropna(axis=0, inplace=True)
        return(data)

    '''
    In this function, 
    The data is preprocessed to be in between 0 and 1, this makes the data sutiable for Recurrent Neural Network
    '''
    def data_scaling(self, dfm):
        scaler = MinMaxScaler(feature_range=(0, 1))
        dg = pd.DataFrame(scaler.fit_transform(dfm[["High", "Low", "Open", "Close", "Volume", "First Day Current Month Opening",
                                                    "mv_avg_12", "mv_avg_24", "First Day Next Month Opening"]].values))
        X = dg[[0, 1, 2, 3, 4, 5, 6, 7]]
        X = self.create_window(X, self.window)
        X = np.reshape(X.values, (X.shape[0], self.window+1, 8))
        y = np.array(dg[8][self.window:])

        return X, y
